// -*- Java -*-
/*!
 * @file  MultipleInterfaceImpl.java
 * @brief ModuleDescription
 * @date  $Date$
 *
 * $Id$
 */

import RTC.TimedShortSeq;
import RTC.TimedDoubleSeq;
import jp.go.aist.rtm.RTC.DataFlowComponentBase;
import jp.go.aist.rtm.RTC.Manager;
import jp.go.aist.rtm.RTC.port.InPort;
import jp.go.aist.rtm.RTC.port.OutPort;
import jp.go.aist.rtm.RTC.util.DataRef;
import RTC.ReturnCode_t;
import RTC.Time;


import javax.vecmath.*;

import java.io.BufferedReader;
import java.io.FileReader;

/*!
 * @class MultipleInterfaceImpl
 * @brief ModuleDescription
 *
 */
public class MultipleInterfaceImpl extends DataFlowComponentBase {
	
	//public value initialize
	int robotNumberPos=5;
	int robotNum=5;
	
	GUIFrame myFrame;
	
	Point3d[] robotPosition=null;
	Point2d[] robotVelocity=null;
	
	public int[][] lrfDistance = null;
	public Point2d[][] lrfDataGlobal=null;
	public Point2d[][] lrfDataRobot=null;

  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
	public MultipleInterfaceImpl(Manager manager) {  
        super(manager);
        // <rtc-template block="initializer">
        m_lrf_val = new TimedShortSeq(new Time(0,0),new short[0]);
        m_lrf = new DataRef<TimedShortSeq>(m_lrf_val);
        m_lrfInIn = new InPort<TimedShortSeq>("lrfIn", m_lrf);
        
        m_pos_val = new TimedDoubleSeq(new Time(0,0),new double[0]);
        m_pos = new DataRef<TimedDoubleSeq>(m_pos_val);
        m_posInIn = new InPort<TimedDoubleSeq>("posIn", m_pos);
        
        m_velIn_val = new TimedDoubleSeq(new Time(0,0),new double[0]);
        m_velIn = new DataRef<TimedDoubleSeq>(m_velIn_val);
        m_velInIn = new InPort<TimedDoubleSeq>("velIn", m_velIn);
        
        m_velOut_val = new TimedDoubleSeq(new Time(0,0),new double[0]);
        m_velOut = new DataRef<TimedDoubleSeq>(m_velOut_val);
        m_velOutOut = new OutPort<TimedDoubleSeq>("velOut", m_velOut);
        
        m_commandOut_val = new TimedDoubleSeq(new Time(0,0),new double[0]);
        m_commandOut = new DataRef<TimedDoubleSeq>(m_commandOut_val);
        m_commandOutOut = new OutPort<TimedDoubleSeq>("commandOut", m_commandOut);
        // </rtc-template>

    }

    /**
     *
     * The initialize action (on CREATED->ALIVE transition)
     * formaer rtc_init_entry() 
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
    @Override
    protected ReturnCode_t onInitialize() {
        // Registration: InPort/OutPort/Service
        // <rtc-template block="registration">
        // Set InPort buffers
        addInPort("lrfIn", m_lrfInIn);
        addInPort("posIn", m_posInIn);
        addInPort("velIn", m_velInIn);
        
        // Set OutPort buffer
        addOutPort("velOut", m_velOutOut);
        addOutPort("commandOut", m_commandOutOut);
        // </rtc-template>
        
        
        
		//設定ファイルからデータを読み取る
		//設定ファイル名
		String filename ="robot.conf";
		//ファイル読込ストリーム
		BufferedReader br;
		//読み取り文字列
		String readString="";
		//行数
		int dataRow=0;

		try{
			//ファイルにアクセス
			br = new BufferedReader(new FileReader(filename));
			while(readString !=null){
				//一行ずつ読み取り
				readString =br.readLine();
				if(readString !=null){
					//Endがあれば終了
					if(readString.equals("End"))	break;
					char[] readChar = readString.toCharArray();
					if(readChar[0] != '%'){	//コメント行をはじく
						//最初の一行はフロントエンドのＩＰアドレス
						if(dataRow==0)	robotNum = Integer.parseInt(readString);

						dataRow++;
					}
				}
			}
		}catch(Exception eFile){
			System.out.println(eFile.toString());
			System.out.println("ロボット台数を5台でスタートします");
			robotNum=5;
		}
		
		//JoyStickの取得開始
//		JoyClass.openDevice();

		//GUI起動
		myFrame=new GUIFrame();
		myFrame.setLocation(480, 0);

		myFrame.mapPanel.robotNum=robotNum;

		myFrame.mapPanel.robotData = new MapRobotData[myFrame.mapPanel.robotNum];
		for(int i=0;i<myFrame.mapPanel.robotNum;i++){
			myFrame.mapPanel.robotData[i] = new MapRobotData();
		}
		for(int i=0;i<robotNum;i++){
			myFrame.mapPanel.robotData[i].posdata=new Point3d((double)myFrame.mapPanel.robotRadius*2.3*(-i),0,0);
		}
		for(int i=0;i<robotNum;i++){
			myFrame.mapPanel.robotData[i].velocity=new Point2d(0,0);
		}

		
		myFrame.setVisible(true);
		
        
        return super.onInitialize();
    }

    /***
     *
     * The finalize action (on ALIVE->END transition)
     * formaer rtc_exiting_entry()
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onFinalize() {
//        return super.onFinalize();
//    }

    /***
     *
     * The startup action when ExecutionContext startup
     * former rtc_starting_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onStartup(int ec_id) {
//        return super.onStartup(ec_id);
//    }

    /***
     *
     * The shutdown action when ExecutionContext stop
     * former rtc_stopping_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onShutdown(int ec_id) {
//        return super.onShutdown(ec_id);
//    }

    /***
     *
     * The activated action (Active state entry action)
     * former rtc_active_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onActivated(int ec_id) {
//        return super.onActivated(ec_id);
//    }

    /***
     *
     * The deactivated action (Active state exit action)
     * former rtc_active_exit()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onDeactivated(int ec_id) {
//        return super.onDeactivated(ec_id);
//    }

    /***
     *
     * The execution action that is invoked periodically
     * former rtc_active_do()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
    @Override
    protected ReturnCode_t onExecute(int ec_id) {
    	
    	if(m_posInIn.isNew()){
//        	System.out.println("pos new");
    		m_posInIn.read();
    		if(m_pos.v.data !=null){
    			robotNumberPos=(m_pos.v.data.length+1)/3;
    			robotNum=robotNumberPos;
    			robotPosition = new Point3d[robotNumberPos];
       			//System.out.println("robotNumberPos:"+robotNumberPos);
       			int robotNumCount=0;
       			double x=0,y=0,z=0;
        		for(int i=0;i<m_pos.v.data.length;i++){
           			//System.out.println("pos["+i+"]:"+m_posSeq.v.data[i]);
           			if(i%3==0)	x=m_pos.v.data[i];
           			if(i%3==1)	y=m_pos.v.data[i];
           			if(i%3==2){
           				z=m_pos.v.data[i];
               			robotPosition[robotNumCount]=new Point3d(x,y,z);
               			robotNumCount++;
           			}
        		}
//        		for(int i=0;i<robotPosition.length;i++){
//        			System.out.println("robotPosition["+i+"]:"+robotPosition[i]);
//        		}
    		}
    	}
    	
    	if(m_velInIn.isNew()){
    		m_velInIn.read();
    		if(m_velIn.v.data !=null){
    			int robotNumberVel=(m_velIn.v.data.length+1)/2;
    			robotVelocity = new Point2d[robotNumberVel];
       			int robotNumVelCount=0;
       			double x=0,y=0;
        		for(int i=0;i<m_velIn.v.data.length;i++){
           			System.out.println("vel["+i+"]:"+m_velIn.v.data[i]);
           			if(i%2==0)	x=m_velIn.v.data[i];
           			if(i%2==1){
           				y=m_velIn.v.data[i];
               			robotVelocity[robotNumVelCount]=new Point2d(x,y);
               			robotNumVelCount++;
           			}
        		}
    		}
    	}
    	
    	
    	if(m_lrfInIn.isNew()){
//        	System.out.println("lrf new");
    		m_lrfInIn.read();
    		if(m_lrf.v.data!=null){
    			int lrfNum=m_lrf.v.data.length/robotNumberPos;
    			lrfDistance=new int[robotNumberPos][lrfNum];
    			System.out.println("lrfDistance.length"+m_lrf.v.data.length+","+lrfNum);    		

    			int robotNumCount=0;
    			int lrfCount=0;
    			
    			for(int i=0;i<m_lrf.v.data.length;i++){
//    				System.out.println("lrf:"+robotNumCount+","+lrfCount+" ("+i+") ->"+m_lrf.v.data[i]);
    				lrfDistance[robotNumCount][lrfCount]= m_lrf.v.data[i];
    				lrfCount++;
    				if(lrfCount==lrfNum){
    					robotNumCount++;
    					lrfCount=0;
    				}
    			}
    			
				//LRFデータをパネルに渡す
				//距離データを座標に変換する
				if(lrfDistance !=null){
					
		    		lrfDataRobot = new Point2d[robotNum][lrfNum];
		    		lrfDataGlobal = new Point2d[robotNum][lrfNum];

					for(int i=0;i<robotNum;i++){
						for(int j=0;j<lrfNum;j++){
							//ロボット座標でのLRFデータ
							//URGの場合
//							double x= lrfDistance[i][j] * Math.cos(Math.PI*1.5/lrfNum*j-Math.PI*3/4);
//			    			double y= lrfDistance[i][j] * Math.sin(Math.PI*1.5/lrfNum*j-Math.PI*3/4);
							
							//シミュレータの場合
							double x= -lrfDistance[i][j] * Math.cos(Math.PI*2.0/lrfNum*j-Math.PI);
			    			double y= -lrfDistance[i][j] * Math.sin(Math.PI*2.0/lrfNum*j-Math.PI);
			    			lrfDataRobot[i][j] = new Point2d(x, y);
			    			
							//グローバル座標に変換
			    			//ロボット座標系からグローバルマップ系に変換
			    			//回転行列
			    			Matrix3d Rz = new Matrix3d(); // マトリクスを作成
			    			Rz.rotZ(-robotPosition[i].z); // 回転マトリクスをセット
			    			//並進ベクトル
			    			Vector3d transV = new Vector3d(robotPosition[i].x, robotPosition[i].y, 0);
			    			//同次変換行列の作成
			    			Matrix4d transAll = new Matrix4d();
			    			transAll.set(Rz, transV, 1);
			    			Vector4d v = new Vector4d(x, y, 0, 1); // ベクトルの初期値
			    			transAll.transform(v);	// TransAll を v に作用させる
			    			lrfDataGlobal[i][j] = new Point2d(v.x, v.y);
						}
					}
				}
    		}
    	}
    	
    	//速度データの出力
//		double vJoy=-JoyClass.axisValue[1]/200.0;
//		double wJoy=-JoyClass.axisValue[2]/200.0;
    	m_velOut_val.data=new double[robotNum*2];
//		for(int i=0;i<m_velOut_val.data.length;i++){
//			if(i%2==0)	m_velOut_val.data[i]=vJoy;
//			if(i%2==1)	m_velOut_val.data[i]=wJoy;
//		}

		
		for(int i=0;i<robotNum;i++){
			if(myFrame.mapPanel.robotData[i]!=null){
				if(myFrame.mapPanel.robotData[i].posdata !=null){
					if(myFrame.mapPanel.groupData !=null){
						if(myFrame.mapPanel.robotData[i].groupID !=-1){
							if(myFrame.mapPanel.groupData[myFrame.mapPanel.robotData[i].groupID] !=null){
								if(myFrame.mapPanel.groupData[myFrame.mapPanel.robotData[i].groupID].targetTrajectry !=null){
									if(robotVelocity[i] !=null){
										VirtualTargetFollowing vtf = new VirtualTargetFollowing();
										robotVelocity[i] = vtf.getVelocity(myFrame.mapPanel.robotData[i].posdata, myFrame.mapPanel.groupData[myFrame.mapPanel.robotData[i].groupID].targetTrajectry, robotVelocity[i], lrfDataRobot[i]);
										//System.out.println("vel in vtf:"+robotVelocity[i]);
									}
								}
							}
						}
					}
				}
			}
		}
		if(robotVelocity !=null){
			int robotVelCount=0;
	    	m_velOut_val.data=new double[robotNum*2];
			for(int i=0;i<m_velOut_val.data.length;i++){
				if(i%2==0){
					if(robotVelocity[robotVelCount] !=null){
						m_velOut_val.data[i]=robotVelocity[robotVelCount].x;
					}
				}
				if(i%2==1){
					if(robotVelocity[robotVelCount] !=null){
						m_velOut_val.data[i]=robotVelocity[robotVelCount].y;
					}
					robotVelCount++;
				}
			}
		}
//		for(int i=0;i<m_velOut_val.data.length;i++){
//			System.out.println("vel["+i+"]:"+m_velOut_val.data[i]);
//		}
		m_velOutOut.write();

    	//パネルへの描画を行う
		myFrame.mapPanel.robotNum=robotNum;
		myFrame.mapPanel.robotNowUrgPoint = new Point2d[robotNum][];
		for(int i=0;i<robotNum;i++){
			//自己位置をシミュレータから渡す
			if(robotPosition != null){
				if(robotPosition[i] != null){
					myFrame.mapPanel.robotData[i].posdata = new Point3d(robotPosition[i].x,robotPosition[i].y,robotPosition[i].z);
				}
			}

			//速度をパネルに渡す
			if(robotVelocity != null){
				if(robotVelocity[i] !=null){
					myFrame.mapPanel.robotData[i].velocity = new Point2d(robotVelocity[i].x, robotVelocity[i].y);
				}
			}

			//LRFをパネルに渡す
			if(lrfDataGlobal != null){
				if(lrfDataGlobal[i] != null){
					myFrame.mapPanel.robotNowUrgPoint[i] = new Point2d[lrfDataGlobal[i].length];
					for(int j=0;j<lrfDataGlobal[i].length;j++){
						if(lrfDataGlobal[i][j] != null){
							myFrame.mapPanel.maxLrfDistance=5600;
							myFrame.mapPanel.robotNowUrgPoint[i][j]= new Point2d(lrfDataGlobal[i][j].x, lrfDataGlobal[i][j].y);
						}
					}
				}
			}
		}

		//インタフェースGUIの表示開始
//		myFrame.setVisible(true);
    	
		myFrame.mapPanel.repaint();

        try {
            Thread.sleep(50);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        
        return super.onExecute(ec_id);
    }

    /***
     *
     * The aborting action when main logic error occurred.
     * former rtc_aborting_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//  @Override
//  public ReturnCode_t onAborting(int ec_id) {
//      return super.onAborting(ec_id);
//  }

    /***
     *
     * The error action in ERROR state
     * former rtc_error_do()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    public ReturnCode_t onError(int ec_id) {
//        return super.onError(ec_id);
//    }

    /***
     *
     * The reset action that is invoked resetting
     * This is same but different the former rtc_init_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onReset(int ec_id) {
//        return super.onReset(ec_id);
//    }

    /***
     *
     * The state update action that is invoked after onExecute() action
     * no corresponding operation exists in OpenRTm-aist-0.2.0
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onStateUpdate(int ec_id) {
//        return super.onStateUpdate(ec_id);
//    }

    /***
     *
     * The action that is invoked when execution context's rate is changed
     * no corresponding operation exists in OpenRTm-aist-0.2.0
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onRateChanged(int ec_id) {
//        return super.onRateChanged(ec_id);
//    }
//
    // DataInPort declaration
    // <rtc-template block="inport_declare">
    protected TimedShortSeq m_lrf_val;
    protected DataRef<TimedShortSeq> m_lrf;
    /*!
     */
    protected InPort<TimedShortSeq> m_lrfInIn;

    protected TimedDoubleSeq m_pos_val;
    protected DataRef<TimedDoubleSeq> m_pos;
    /*!
     */
    protected InPort<TimedDoubleSeq> m_posInIn;

    protected TimedDoubleSeq m_velIn_val;
    protected DataRef<TimedDoubleSeq> m_velIn;
    /*!
     */
    protected InPort<TimedDoubleSeq> m_velInIn;

    
    // </rtc-template>

    // DataOutPort declaration
    // <rtc-template block="outport_declare">
    protected TimedDoubleSeq m_velOut_val;
    protected DataRef<TimedDoubleSeq> m_velOut;
    /*!
     */
    protected OutPort<TimedDoubleSeq> m_velOutOut;

    protected TimedDoubleSeq m_commandOut_val;
    protected DataRef<TimedDoubleSeq> m_commandOut;
    /*!
     */
    protected OutPort<TimedDoubleSeq> m_commandOutOut;

    
    // </rtc-template>

    // CORBA Port declaration
    // <rtc-template block="corbaport_declare">
    
    // </rtc-template>

    // Service declaration
    // <rtc-template block="service_declare">
    
    // </rtc-template>

    // Consumer declaration
    // <rtc-template block="consumer_declare">
    
    // </rtc-template>


}
